package geniusweb.exampleparties.timedependentparty;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Collections;
import java.util.LinkedList;

import geniusweb.bidspace.BidsWithUtility;
import geniusweb.bidspace.Interval;
import geniusweb.bidspace.IssueInfo;
import geniusweb.bidspace.SumOfGroupsLinAdditiveAdapter;
import geniusweb.issuevalue.Bid;
import geniusweb.issuevalue.Value;
import geniusweb.profile.utilityspace.LinearAdditive;
import geniusweb.profile.utilityspace.SumOfGroupsUtilitySpace;
import geniusweb.profile.utilityspace.UtilitySpace;
import tudelft.utilities.immutablelist.ImmutableList;
import tudelft.utilities.immutablelist.MapList;

/**
 * Inner class for TimeDependentParty, Handles mechanisms for searching the bid
 * space, and using adapters to handle the nonlinear
 * {@link SumOfGroupsUtilitySpace}.
 * 
 * This class is public for testing purposes. This class may change in the
 * future, use at your own risk.
 *
 */
public class ExtendedUtilSpace {
	private UtilitySpace utilspace;
	// linear approximation of utilspace (possibly utilspace itself)
	// thsi may have different utilityes issues and values than utilspace.
	private LinearAdditive linapprox;

	// bidutils tolerance minUtil maxUtil all work on linapprox
	private BidsWithUtility bidutils;
	/**
	 * The tolerance for a utility. Generally utilities can be this amount
	 * smaller than requested. See also {@link #computeTolerance()}.
	 */
	private BigDecimal tolerance;
	// min and max achievable utility
	private BigDecimal minUtil;
	private BigDecimal maxUtil;

	public ExtendedUtilSpace(UtilitySpace space) {
		this.utilspace = space;
		if (space instanceof LinearAdditive)
			linapprox = (LinearAdditive) space;
		else
			linapprox = SumOfGroupsLinAdditiveAdapter
					.create((SumOfGroupsUtilitySpace) space);
		bidutils = new BidsWithUtility(linapprox);
		computeMinMax();
		this.tolerance = computeTolerance();
	}

	/**
	 * Computes the fields minutil and maxUtil.
	 * <p>
	 * TODO this is simplistic, very expensive method and may cause us to run
	 * out of time on large domains.
	 * <p>
	 * Assumes that utilspace and bidutils have been set properly.
	 */
	private void computeMinMax() {
		Interval range = bidutils.getRange();
		this.minUtil = range.getMin();
		this.maxUtil = range.getMax();

		Bid rvbid = linapprox.getReservationBid();
		if (rvbid != null) {
			BigDecimal rv = linapprox.getUtility(rvbid);
			if (rv.compareTo(minUtil) > 0)
				minUtil = rv;
		}
	}

	/**
	 * Tolerance is the Interval we need when searching bids. When we are close
	 * to the maximum utility, this value has to be the distance between the
	 * best and one-but-best utility.
	 * 
	 * @return the minimum tolerance required in the {@link #linapprox} space,
	 *         which is the minimum difference between the weighted utility of
	 *         the best and one-but-best issue value.
	 */
	protected BigDecimal computeTolerance() {
		BigDecimal tolerance = BigDecimal.ONE;
		for (IssueInfo iss : bidutils.getInfo()) {
			if (iss.getValues().size().compareTo(BigInteger.ONE) > 0) {
				// we have at least 2 values.
				LinkedList<BigDecimal> values = new LinkedList<BigDecimal>();
				for (Value val : iss.getValues()) {
					values.add(iss.getWeightedUtil(val));
				}
				Collections.sort(values);
				Collections.reverse(values);
				tolerance = tolerance
						.min(values.get(0).subtract(values.get(1)));
			}
		}
		return tolerance;
	}

	/**
	 * @return the minimum utility in the {@link #linapprox} space
	 */
	public BigDecimal getMin() {
		return minUtil;
	}

	/**
	 * @return the maximum utility in the {@link #linapprox} space
	 */
	public BigDecimal getMax() {
		return maxUtil;
	}

	/**
	 * @param utilityGoal the requested utility in the {@link #linapprox} space
	 * @return bids with utility inside [utilitygoal-{@link #tolerance},
	 *         utilitygoal], converted back to bids in the utilspace
	 */
	public ImmutableList<Bid> getBids(BigDecimal utilityGoal) {
		ImmutableList<Bid> bidslist = bidutils.getBids(
				new Interval(utilityGoal.subtract(tolerance), utilityGoal));
		return new MapList<>((Bid b) -> realBid(b), bidslist);
	}

	/**
	 * @param b a bid in the linapprox space
	 * @return a bid in the original space
	 */
	private Bid realBid(Bid b) {
		if (linapprox instanceof SumOfGroupsLinAdditiveAdapter)
			return ((SumOfGroupsLinAdditiveAdapter) linapprox).toGroupBid(b);
		return b;
	}

}
